@@ -38,6 +38,7 @@
 #include "stats.h"
 #include "print.h"
 #include "tlv.h"
+#include "tsproc.h"
 #include "uds.h"
 #include "util.h"
 
@@ -72,6 +73,7 @@ struct clock_subscriber {
 };
 
 struct clock {
+	struct config *config;
 	clockid_t clkid;
 	struct servo *servo;
 	enum servo_type servo_type;
@@ -102,14 +104,11 @@ struct clock {
 	enum servo_state servo_state;
 	tmv_t master_offset;
 	tmv_t path_delay;
-	struct filter *delay_filter;
+	tmv_t ingress_ts;
+	struct tsproc *tsproc;
 	struct freq_estimator fest;
 	struct time_status_np status;
 	double nrr;
-	tmv_t c1;
-	tmv_t c2;
-	tmv_t t1;
-	tmv_t t2;
 	struct clock_description desc;
 	struct clock_stats stats;
 	int stats_interval;
@@ -273,7 +272,7 @@ void clock_destroy(struct clock *c)
 		phc_close(c->clkid);
 	}
 	servo_destroy(c->servo);
-	filter_destroy(c->delay_filter);
+	tsproc_destroy(c->tsproc);
 	stats_destroy(c->stats.offset);
 	stats_destroy(c->stats.freq);
 	stats_destroy(c->stats.delay);
@@ -415,7 +414,7 @@ static int clock_management_fill_response(struct clock *c, struct port *p,
 	case TLV_TIME_STATUS_NP:
 		tsn = (struct time_status_np *) tlv->data;
 		tsn->master_offset = c->master_offset;
-		tsn->ingress_time = tmv_to_nanoseconds(c->t2);
+		tsn->ingress_time = tmv_to_nanoseconds(c->ingress_ts);
 		tsn->cumulativeScaledRateOffset =
 			(Integer32) (c->status.cumulativeScaledRateOffset +
 				      c->nrr * POW2_41 - POW2_41);
@@ -484,12 +483,25 @@ static int clock_management_set(struct clock *c, struct port *p,
 {
 	int respond = 0;
 	struct management_tlv *tlv;
+	struct management_tlv_datum *mtd;
 	struct grandmaster_settings_np *gsn;
 	struct subscribe_events_np *sen;
 
 	tlv = (struct management_tlv *) req->management.suffix;
 
 	switch (id) {
+	case TLV_PRIORITY1:
+		mtd = (struct management_tlv_datum *) tlv->data;
+		c->dds.priority1 = mtd->val;
+		*changed = 1;
+		respond = 1;
+		break;
+	case TLV_PRIORITY2:
+		mtd = (struct management_tlv_datum *) tlv->data;
+		c->dds.priority2 = mtd->val;
+		*changed = 1;
+		respond = 1;
+		break;
 	case TLV_GRANDMASTER_SETTINGS_NP:
 		gsn = (struct grandmaster_settings_np *) tlv->data;
 		c->dds.clockQuality = gsn->clockQuality;
@@ -545,20 +557,14 @@ static void clock_stats_update(struct clock_stats *s,
 	stats_reset(s->delay);
 }
 
-static enum servo_state clock_no_adjust(struct clock *c)
+static enum servo_state clock_no_adjust(struct clock *c, tmv_t ingress,
+					tmv_t origin)
 {
 	double fui;
 	double ratio, freq;
-	tmv_t origin2;
 	struct freq_estimator *f = &c->fest;
 	enum servo_state state = SERVO_UNLOCKED;
 	/*
-	 * We have clock.t1 as the origin time stamp, and clock.t2 as
-	 * the ingress. According to the master's clock, the time at
-	 * which the sync arrived is:
-	 *
-	 *    origin = origin_ts + path_delay + correction
-	 *
 	 * The ratio of the local clock freqency to the master clock
 	 * is estimated by:
 	 *
@@ -570,25 +576,21 @@ static enum servo_state clock_no_adjust(struct clock *c)
 	 * error caused by our imperfect path delay measurement.
 	 */
 	if (!f->ingress1) {
-		f->ingress1 = c->t2;
-		f->origin1 = tmv_add(c->t1, tmv_add(c->c1, c->c2));
+		f->ingress1 = ingress;
+		f->origin1 = origin;
 		return state;
 	}
 	f->count++;
 	if (f->count < f->max_count) {
 		return state;
 	}
-	if (tmv_eq(c->t2, f->ingress1)) {
+	if (tmv_eq(ingress, f->ingress1)) {
 		pr_warning("bad timestamps in rate ratio calculation");
 		return state;
 	}
-	/*
-	 * origin2 = c->t1 (+c->path_delay) + c->c1 + c->c2;
-	 */
-	origin2 = tmv_add(c->t1, tmv_add(c->c1, c->c2));
 
-	ratio = tmv_dbl(tmv_sub(origin2, f->origin1)) /
-		tmv_dbl(tmv_sub(c->t2, f->ingress1));
+	ratio = tmv_dbl(tmv_sub(origin, f->origin1)) /
+		tmv_dbl(tmv_sub(ingress, f->ingress1));
 	freq = (1.0 - ratio) * 1e9;
 
 	if (c->stats.max_count > 1) {
@@ -610,8 +612,8 @@ static enum servo_state clock_no_adjust(struct clock *c)
 	pr_debug("master/local  %.9f", ratio);
 	pr_debug("diff         %+.9f", ratio - (fui + c->nrr - 1.0));
 
-	f->ingress1 = c->t2;
-	f->origin1 = origin2;
+	f->ingress1 = ingress;
+	f->origin1 = origin;
 	f->count = 0;
 
 	return state;
@@ -753,6 +755,11 @@ UInteger8 clock_class(struct clock *c)
 	return c->dds.clockQuality.clockClass;
 }
 
+struct config *clock_config(struct clock *c)
+{
+	return c->config;
+}
+
 static int clock_add_port(struct clock *c, int phc_index,
 			  enum timestamp_type timestamping,
 			  struct interface *iface)
@@ -791,17 +798,21 @@ static void clock_remove_port(struct clock *c, struct port *p)
 	port_close(p);
 }
 
-struct clock *clock_create(int phc_index, struct interfaces_head *ifaces,
-			   enum timestamp_type timestamping, struct default_ds *dds,
-			   enum servo_type servo)
+struct clock *clock_create(struct config *config, int phc_index,
+			   struct interfaces_head *ifaces,
+			   struct default_ds *dds)
 {
+	enum timestamp_type timestamping =
+		config_get_int(config, NULL, "time_stamping");
 	int fadj = 0, max_adj = 0, sw_ts = timestamping == TS_SOFTWARE ? 1 : 0;
+	enum servo_type servo = config_get_int(config, NULL, "clock_servo");
 	struct clock *c = &the_clock;
 	struct port *p;
 	char phc[32];
 	struct interface *udsif = &c->uds_interface;
 	struct interface *iface;
 	struct timespec ts;
+	int sfl;
 
 	clock_gettime(CLOCK_REALTIME, &ts);
 	srandom(ts.tv_sec ^ ts.tv_nsec);
@@ -809,16 +820,27 @@ struct clock *clock_create(int phc_index, struct interfaces_head *ifaces,
 	if (c->nports)
 		clock_destroy(c);
 
-	snprintf(udsif->name, sizeof(udsif->name), "%s", uds_path);
-	udsif->transport = TRANS_UDS;
-	udsif->delay_filter_length = 1;
+	snprintf(udsif->name, sizeof(udsif->name), "%s",
+		 config_get_string(config, NULL, "uds_address"));
+	if (config_set_section_int(config, udsif->name,
+				    "delay_mechanism", DM_AUTO)) {
+		return NULL;
+	}
+	if (config_set_section_int(config, udsif->name,
+				    "network_transport", TRANS_UDS)) {
+		return NULL;
+	}
+	if (config_set_section_int(config, udsif->name, "delay_filter_length", 1)) {
+		return NULL;
+	}
 
-	c->free_running = dds->free_running;
-	c->freq_est_interval = dds->freq_est_interval;
-	c->grand_master_capable = dds->grand_master_capable;
-	c->kernel_leap = dds->kernel_leap;
+	c->config = config;
+	c->free_running = config_get_int(config, NULL, "free_running");
+	c->freq_est_interval = config_get_int(config, NULL, "freq_est_interval");
+	c->grand_master_capable = config_get_int(config, NULL, "gmCapable");
+	c->kernel_leap = config_get_int(config, NULL, "kernel_leap");
 	c->utc_offset = CURRENT_UTC_OFFSET;
-	c->time_source = dds->time_source;
+	c->time_source = config_get_int(config, NULL, "timeSource");
 	c->desc = dds->clock_desc;
 
 	if (c->free_running) {
@@ -857,21 +879,22 @@ struct clock *clock_create(int phc_index, struct interfaces_head *ifaces,
 		   the actual frequency of the clock. */
 		clockadj_set_freq(c->clkid, fadj);
 	}
-	c->servo = servo_create(servo, -fadj, max_adj, sw_ts);
+	c->servo = servo_create(c->config, servo, -fadj, max_adj, sw_ts);
 	if (!c->servo) {
 		pr_err("Failed to create clock servo");
 		return NULL;
 	}
 	c->servo_state = SERVO_UNLOCKED;
 	c->servo_type = servo;
-	c->delay_filter = filter_create(dds->delay_filter,
-					dds->delay_filter_length);
-	if (!c->delay_filter) {
-		pr_err("Failed to create delay filter");
+	c->tsproc = tsproc_create(config_get_int(config, NULL, "tsproc_mode"),
+				  config_get_int(config, NULL, "delay_filter"),
+				  config_get_int(config, NULL, "delay_filter_length"));
+	if (!c->tsproc) {
+		pr_err("Failed to create time stamp processor");
 		return NULL;
 	}
 	c->nrr = 1.0;
-	c->stats_interval = dds->stats_interval;
+	c->stats_interval = config_get_int(config, NULL, "summary_interval");
 	c->stats.offset = stats_create();
 	c->stats.freq = stats_create();
 	c->stats.delay = stats_create();
@@ -879,8 +902,9 @@ struct clock *clock_create(int phc_index, struct interfaces_head *ifaces,
 		pr_err("failed to create stats");
 		return NULL;
 	}
-	if (dds->sanity_freq_limit) {
-		c->sanity_check = clockcheck_create(dds->sanity_freq_limit);
+	sfl = config_get_int(config, NULL, "sanity_freq_limit");
+	if (sfl) {
+		c->sanity_check = clockcheck_create(sfl);
 		if (!c->sanity_check) {
 			pr_err("Failed to create clock sanity check");
 			return NULL;
@@ -1289,63 +1313,29 @@ int clock_poll(struct clock *c)
 	return 0;
 }
 
-void clock_path_delay(struct clock *c, struct timespec req, struct timestamp rx,
-		      Integer64 correction)
+void clock_path_delay(struct clock *c, tmv_t req, tmv_t rx)
 {
-	tmv_t c1, c2, c3, pd, t1, t2, t3, t4;
-	double rr;
+	tsproc_up_ts(c->tsproc, req, rx);
 
-	if (tmv_is_zero(c->t1))
+	if (tsproc_update_delay(c->tsproc, &c->path_delay))
 		return;
 
-	c1 = c->c1;
-	c2 = c->c2;
-	c3 = correction_to_tmv(correction);
-	t1 = c->t1;
-	t2 = c->t2;
-	t3 = timespec_to_tmv(req);
-	t4 = timestamp_to_tmv(rx);
-	rr = clock_rate_ratio(c);
-
-	/*
-	 * c->path_delay = (t2 - t3) * rr + (t4 - t1);
-	 * c->path_delay -= c_sync + c_fup + c_delay_resp;
-	 * c->path_delay /= 2.0;
-	 */
-
-	pd = tmv_sub(t2, t3);
-	if (rr != 1.0)
-		pd = dbl_tmv(tmv_dbl(pd) * rr);
-	pd = tmv_add(pd, tmv_sub(t4, t1));
-	pd = tmv_sub(pd, tmv_add(c1, tmv_add(c2, c3)));
-	pd = tmv_div(pd, 2);
-
-	if (pd < 0) {
-		pr_debug("negative path delay %10" PRId64, pd);
-		pr_debug("path_delay = (t2 - t3) * rr + (t4 - t1) - (c1 + c2 + c3)");
-		pr_debug("t2 - t3 = %+10" PRId64, t2 - t3);
-		pr_debug("t4 - t1 = %+10" PRId64, t4 - t1);
-		pr_debug("rr = %.9f", rr);
-		pr_debug("c1 %10" PRId64, c1);
-		pr_debug("c2 %10" PRId64, c2);
-		pr_debug("c3 %10" PRId64, c3);
-	}
-
-	c->path_delay = filter_sample(c->delay_filter, pd);
-
 	c->cur.meanPathDelay = tmv_to_TimeInterval(c->path_delay);
 
-	pr_debug("path delay    %10" PRId64 " %10" PRId64, c->path_delay, pd);
-
 	if (c->stats.delay)
-		stats_add_value(c->stats.delay, tmv_to_nanoseconds(pd));
+		stats_add_value(c->stats.delay,
+				tmv_to_nanoseconds(c->path_delay));
 }
 
-void clock_peer_delay(struct clock *c, tmv_t ppd, double nrr)
+void clock_peer_delay(struct clock *c, tmv_t ppd, tmv_t req, tmv_t rx,
+		      double nrr)
 {
 	c->path_delay = ppd;
 	c->nrr = nrr;
 
+	tsproc_set_delay(c->tsproc, ppd);
+	tsproc_up_ts(c->tsproc, req, rx);
+
 	if (c->stats.delay)
 		stats_add_value(c->stats.delay, tmv_to_nanoseconds(ppd));
 }
@@ -1381,7 +1371,7 @@ int clock_switch_phc(struct clock *c, int phc_index)
 	}
 	fadj = (int) clockadj_get_freq(clkid);
 	clockadj_set_freq(clkid, fadj);
-	servo = servo_create(c->servo_type, -fadj, max_adj, 0);
+	servo = servo_create(c->config, c->servo_type, -fadj, max_adj, 0);
 	if (!servo) {
 		pr_err("Switching PHC, failed to create clock servo");
 		phc_close(clkid);
@@ -1395,32 +1385,16 @@ int clock_switch_phc(struct clock *c, int phc_index)
 	return 0;
 }
 
-enum servo_state clock_synchronize(struct clock *c,
-				   struct timespec ingress_ts,
-				   struct timestamp origin_ts,
-				   Integer64 correction1,
-				   Integer64 correction2)
+enum servo_state clock_synchronize(struct clock *c, tmv_t ingress, tmv_t origin)
 {
-	double adj;
-	tmv_t ingress, origin;
+	double adj, weight;
 	enum servo_state state = SERVO_UNLOCKED;
 
-	ingress = timespec_to_tmv(ingress_ts);
-	origin  = timestamp_to_tmv(origin_ts);
+	c->ingress_ts = ingress;
 
-	c->t1 = origin;
-	c->t2 = ingress;
+	tsproc_down_ts(c->tsproc, origin, ingress);
 
-	c->c1 = correction_to_tmv(correction1);
-	c->c2 = correction_to_tmv(correction2);
-
-	/*
-	 * c->master_offset = ingress - origin - c->path_delay - c->c1 - c->c2;
-	 */
-	c->master_offset = tmv_sub(ingress,
-		tmv_add(origin, tmv_add(c->path_delay, tmv_add(c->c1, c->c2))));
-
-	if (!c->path_delay)
+	if (tsproc_update_offset(c->tsproc, &c->master_offset, &weight))
 		return state;
 
 	if (clock_utc_correct(c, ingress))
@@ -1429,10 +1403,10 @@ enum servo_state clock_synchronize(struct clock *c,
 	c->cur.offsetFromMaster = tmv_to_TimeInterval(c->master_offset);
 
 	if (c->free_running)
-		return clock_no_adjust(c);
+		return clock_no_adjust(c, ingress, origin);
 
 	adj = servo_sample(c->servo, tmv_to_nanoseconds(c->master_offset),
-			   tmv_to_nanoseconds(ingress), &state);
+			   tmv_to_nanoseconds(ingress), weight, &state);
 	c->servo_state = state;
 
 	if (c->stats.max_count > 1) {
@@ -1445,19 +1419,21 @@ enum servo_state clock_synchronize(struct clock *c,
 			tmv_to_nanoseconds(c->path_delay));
 	}
 
+	tsproc_set_clock_rate_ratio(c->tsproc, clock_rate_ratio(c));
+
 	switch (state) {
 	case SERVO_UNLOCKED:
 		break;
 	case SERVO_JUMP:
 		clockadj_set_freq(c->clkid, -adj);
 		clockadj_step(c->clkid, -tmv_to_nanoseconds(c->master_offset));
-		c->t1 = tmv_zero();
-		c->t2 = tmv_zero();
+		c->ingress_ts = tmv_zero();
 		if (c->sanity_check) {
 			clockcheck_set_freq(c->sanity_check, -adj);
 			clockcheck_step(c->sanity_check,
 					-tmv_to_nanoseconds(c->master_offset));
 		}
+		tsproc_reset(c->tsproc, 0);
 		break;
 	case SERVO_LOCKED:
 		clockadj_set_freq(c->clkid, -adj);
@@ -1531,9 +1507,8 @@ static void handle_state_decision_event(struct clock *c)
 
 	if (!cid_eq(&best_id, &c->best_id)) {
 		clock_freq_est_reset(c);
-		filter_reset(c->delay_filter);
-		c->t1 = tmv_zero();
-		c->t2 = tmv_zero();
+		tsproc_reset(c->tsproc, 1);
+		c->ingress_ts = tmv_zero();
 		c->path_delay = 0;
 		c->nrr = 1.0;
 		fresh_best = 1;
