@@ -35,64 +35,15 @@
 struct interface {
 	STAILQ_ENTRY(interface) list;
 	char name[MAX_IFNAME_SIZE + 1];
-	enum delay_mechanism dm;
-	enum transport_type transport;
-	struct port_defaults pod;
 	struct sk_ts_info ts_info;
-	enum filter_type delay_filter;
-	int delay_filter_length;
-	int boundary_clock_jbod;
 };
 
-#define CFG_IGNORE_DM           (1 << 0)
-#define CFG_IGNORE_TRANSPORT    (1 << 1)
-#define CFG_IGNORE_TIMESTAMPING (1 << 2)
-#define CFG_IGNORE_SLAVEONLY    (1 << 3)
-#define CFG_IGNORE_PRINT_LEVEL  (1 << 4)
-#define CFG_IGNORE_USE_SYSLOG   (1 << 5)
-#define CFG_IGNORE_VERBOSE      (1 << 6)
-
 struct config {
-	/* configuration override */
-	int cfg_ignore;
-
 	/* configured interfaces */
 	STAILQ_HEAD(interfaces_head, interface) interfaces;
 
-	enum timestamp_type timestamping;
-	enum transport_type transport;
-	enum delay_mechanism dm;
-
-	struct default_ds dds;
-	struct port_defaults pod;
-	int *assume_two_step;
-	int *tx_timestamp_timeout;
-	int *check_fup_sync;
-
-	enum servo_type clock_servo;
-
-	double *step_threshold;
-	double *first_step_threshold;
-	int *max_frequency;
-
-	double *pi_proportional_const;
-	double *pi_integral_const;
-	double *pi_proportional_scale;
-	double *pi_proportional_exponent;
-	double *pi_proportional_norm_max;
-	double *pi_integral_scale;
-	double *pi_integral_exponent;
-	double *pi_integral_norm_max;
-	int *ntpshm_segment;
-
-	unsigned char *ptp_dst_mac;
-	unsigned char *p2p_dst_mac;
-	unsigned char *udp6_scope;
-	char *uds_address;
-
-	int print_level;
-	int use_syslog;
-	int verbose;
+	/* hash of all non-legacy items */
+	struct hash *htab;
 };
 
 int config_read(char *name, struct config *cfg);
@@ -100,4 +51,31 @@ struct interface *config_create_interface(char *name, struct config *cfg);
 void config_init_interface(struct interface *iface, struct config *cfg);
 void config_destroy(struct config *cfg);
 
+/* New, hash table based methods: */
+
+struct config *config_create(void);
+
+double config_get_double(struct config *cfg, const char *section,
+			 const char *option);
+
+int config_get_int(struct config *cfg, const char *section,
+		   const char *option);
+
+char *config_get_string(struct config *cfg, const char *section,
+			const char *option);
+
+int config_set_double(struct config *cfg, const char *option, double val);
+
+int config_set_section_int(struct config *cfg, const char *section,
+			   const char *option, int val);
+
+static inline int config_set_int(struct config *cfg,
+				 const char *option, int val)
+{
+	return config_set_section_int(cfg, NULL, option, val);
+}
+
+int config_set_string(struct config *cfg, const char *option,
+		      const char *val);
+
 #endif
