@@ -39,99 +39,7 @@
 
 int assume_two_step = 0;
 
-static struct config cfg_settings = {
-	.interfaces = STAILQ_HEAD_INITIALIZER(cfg_settings.interfaces),
-
-	.dds = {
-		.dds = {
-			.flags = DDS_TWO_STEP_FLAG,
-			.priority1 = 128,
-			.clockQuality.clockClass = 248,
-			.clockQuality.clockAccuracy = 0xfe,
-			.clockQuality.offsetScaledLogVariance = 0xffff,
-			.priority2 = 128,
-			.domainNumber = 0,
-		},
-		.free_running = 0,
-		.freq_est_interval = 1,
-		.grand_master_capable = 1,
-		.stats_interval = 0,
-		.kernel_leap = 1,
-		.sanity_freq_limit = 200000000,
-		.time_source = INTERNAL_OSCILLATOR,
-		.clock_desc = {
-			.productDescription = {
-				.max_symbols = 64,
-				.text = ";;",
-				.length = 2,
-			},
-			.revisionData = {
-				.max_symbols = 32,
-				.text = ";;",
-				.length = 2,
-			},
-			.userDescription      = { .max_symbols = 128 },
-			.manufacturerIdentity = { 0, 0, 0 },
-		},
-		.delay_filter = FILTER_MOVING_MEDIAN,
-		.delay_filter_length = 10,
-		.boundary_clock_jbod = 0,
-	},
-
-	.pod = {
-		.logAnnounceInterval = 1,
-		.logSyncInterval = 0,
-		.logMinDelayReqInterval = 0,
-		.logMinPdelayReqInterval = 0,
-		.announceReceiptTimeout = 3,
-		.syncReceiptTimeout = 0,
-		.transportSpecific = 0,
-		.announce_span = 1,
-		.path_trace_enabled = 0,
-		.follow_up_info = 0,
-		.freq_est_interval = 1,
-		/* Default to very a large neighborPropDelay threshold */
-		.neighborPropDelayThresh = 20000000,
-		.min_neighbor_prop_delay = -20000000,
-		.tx_timestamp_offset = 0,
-		.rx_timestamp_offset = 0,
-	},
-
-	.timestamping = TS_HARDWARE,
-	.dm = DM_E2E,
-	.transport = TRANS_UDP_IPV4,
-
-	.assume_two_step = &assume_two_step,
-	.tx_timestamp_timeout = &sk_tx_timeout,
-	.check_fup_sync = &sk_check_fupsync,
-
-	.clock_servo = CLOCK_SERVO_PI,
-
-	.step_threshold = &servo_step_threshold,
-	.first_step_threshold = &servo_first_step_threshold,
-	.max_frequency = &servo_max_frequency,
-
-	.pi_proportional_const = &configured_pi_kp,
-	.pi_integral_const = &configured_pi_ki,
-	.pi_proportional_scale = &configured_pi_kp_scale,
-	.pi_proportional_exponent = &configured_pi_kp_exponent,
-	.pi_proportional_norm_max = &configured_pi_kp_norm_max,
-	.pi_integral_scale = &configured_pi_ki_scale,
-	.pi_integral_exponent = &configured_pi_ki_exponent,
-	.pi_integral_norm_max = &configured_pi_ki_norm_max,
-	.ntpshm_segment = &ntpshm_segment,
-
-	.ptp_dst_mac = ptp_dst_mac,
-	.p2p_dst_mac = p2p_dst_mac,
-	.udp6_scope = &udp6_scope,
-	.uds_address = uds_path,
-
-	.print_level = LOG_INFO,
-	.use_syslog = 1,
-	.verbose = 0,
-
-	.cfg_ignore = 0,
-};
+static struct default_ds ptp4l_dds;
 
 static void usage(char *progname)
 {
@@ -167,24 +75,22 @@ static void usage(char *progname)
 
 int main(int argc, char *argv[])
 {
-	char *config = NULL, *req_phc = NULL, *progname;
-	int c, i;
+	char *config = NULL, *req_phc = NULL, *progname, *tmp;
+	int c, err = -1;
 	struct interface *iface;
-	int *cfg_ignore = &cfg_settings.cfg_ignore;
-	enum delay_mechanism *dm = &cfg_settings.dm;
-	enum transport_type *transport = &cfg_settings.transport;
-	enum timestamp_type *timestamping = &cfg_settings.timestamping;
-	struct clock *clock;
-	struct defaultDS *ds = &cfg_settings.dds.dds;
-	int phc_index = -1, required_modes = 0;
+	struct clock *clock = NULL;
+	struct config *cfg;
+	struct default_ds *dds = &ptp4l_dds;
+	struct defaultDS *ds = &ptp4l_dds.dds;
+	int phc_index = -1, print_level, required_modes = 0;
+	unsigned char oui[OUI_LEN];
 
 	if (handle_term_signals())
 		return -1;
 
-	/* Set fault timeouts to a default value */
-	for (i = 0; i < FT_CNT; i++) {
-		cfg_settings.pod.flt_interval_pertype[i].type = FTMO_LOG2_SECONDS;
-		cfg_settings.pod.flt_interval_pertype[i].val = 4;
+	cfg = config_create();
+	if (!cfg) {
+		return -1;
 	}
 
 	/* Process the command line arguments. */
@@ -193,68 +99,70 @@ int main(int argc, char *argv[])
 	while (EOF != (c = getopt(argc, argv, "AEP246HSLf:i:p:sl:mqvh"))) {
 		switch (c) {
 		case 'A':
-			*dm = DM_AUTO;
-			*cfg_ignore |= CFG_IGNORE_DM;
+			if (config_set_int(cfg, "delay_mechanism", DM_AUTO))
+				goto out;
 			break;
 		case 'E':
-			*dm = DM_E2E;
-			*cfg_ignore |= CFG_IGNORE_DM;
+			if (config_set_int(cfg, "delay_mechanism", DM_E2E))
+				goto out;
 			break;
 		case 'P':
-			*dm = DM_P2P;
-			*cfg_ignore |= CFG_IGNORE_DM;
+			if (config_set_int(cfg, "delay_mechanism", DM_P2P))
+				goto out;
 			break;
 		case '2':
-			*transport = TRANS_IEEE_802_3;
-			*cfg_ignore |= CFG_IGNORE_TRANSPORT;
+			if (config_set_int(cfg, "network_transport",
+					    TRANS_IEEE_802_3))
+				goto out;
 			break;
 		case '4':
-			*transport = TRANS_UDP_IPV4;
-			*cfg_ignore |= CFG_IGNORE_TRANSPORT;
+			if (config_set_int(cfg, "network_transport",
+					    TRANS_UDP_IPV4))
+				goto out;
 			break;
 		case '6':
-			*transport = TRANS_UDP_IPV6;
-			*cfg_ignore |= CFG_IGNORE_TRANSPORT;
+			if (config_set_int(cfg, "network_transport",
+					    TRANS_UDP_IPV6))
+				goto out;
 			break;
 		case 'H':
-			*timestamping = TS_HARDWARE;
-			*cfg_ignore |= CFG_IGNORE_TIMESTAMPING;
+			if (config_set_int(cfg, "time_stamping", TS_HARDWARE))
+				goto out;
 			break;
 		case 'S':
-			*timestamping = TS_SOFTWARE;
-			*cfg_ignore |= CFG_IGNORE_TIMESTAMPING;
+			if (config_set_int(cfg, "time_stamping", TS_SOFTWARE))
+				goto out;
 			break;
 		case 'L':
-			*timestamping = TS_LEGACY_HW;
-			*cfg_ignore |= CFG_IGNORE_TIMESTAMPING;
+			if (config_set_int(cfg, "time_stamping", TS_LEGACY_HW))
+				goto out;
 			break;
 		case 'f':
 			config = optarg;
 			break;
 		case 'i':
-			if (!config_create_interface(optarg, &cfg_settings))
-				return -1;
+			if (!config_create_interface(optarg, cfg))
+				goto out;
 			break;
 		case 'p':
 			req_phc = optarg;
 			break;
 		case 's':
-			ds->flags |= DDS_SLAVE_ONLY;
-			*cfg_ignore |= CFG_IGNORE_SLAVEONLY;
+			if (config_set_int(cfg, "slaveOnly", 1)) {
+				goto out;
+			}
 			break;
 		case 'l':
-			if (get_arg_val_i(c, optarg, &cfg_settings.print_level,
+			if (get_arg_val_i(c, optarg, &print_level,
 					  PRINT_LEVEL_MIN, PRINT_LEVEL_MAX))
-				return -1;
-			*cfg_ignore |= CFG_IGNORE_PRINT_LEVEL;
+				goto out;
+			config_set_int(cfg, "logging_level", print_level);
 			break;
 		case 'm':
-			cfg_settings.verbose = 1;
-			*cfg_ignore |= CFG_IGNORE_VERBOSE;
+			config_set_int(cfg, "verbose", 1);
 			break;
 		case 'q':
-			cfg_settings.use_syslog = 0;
-			*cfg_ignore |= CFG_IGNORE_USE_SYSLOG;
+			config_set_int(cfg, "use_syslog", 0);
 			break;
 		case 'v':
 			version_show(stdout);
@@ -264,58 +172,109 @@ int main(int argc, char *argv[])
 			return 0;
 		case '?':
 			usage(progname);
-			return -1;
+			goto out;
 		default:
 			usage(progname);
-			return -1;
+			goto out;
 		}
 	}
 
-	if (config && (c = config_read(config, &cfg_settings))) {
+	if (config && (c = config_read(config, cfg))) {
 		return c;
 	}
-	if (!cfg_settings.dds.grand_master_capable &&
-	    ds->flags & DDS_SLAVE_ONLY) {
-		fprintf(stderr,
-			"Cannot mix 1588 slaveOnly with 802.1AS !gmCapable.\n");
-		return -1;
-	}
-	if (!cfg_settings.dds.grand_master_capable ||
-	    ds->flags & DDS_SLAVE_ONLY) {
-		ds->clockQuality.clockClass = 255;
-	}
-	if (cfg_settings.clock_servo == CLOCK_SERVO_NTPSHM) {
-		cfg_settings.dds.kernel_leap = 0;
-		cfg_settings.dds.sanity_freq_limit = 0;
-	}
 
 	print_set_progname(progname);
-	print_set_verbose(cfg_settings.verbose);
-	print_set_syslog(cfg_settings.use_syslog);
-	print_set_level(cfg_settings.print_level);
+	print_set_verbose(config_get_int(cfg, NULL, "verbose"));
+	print_set_syslog(config_get_int(cfg, NULL, "use_syslog"));
+	print_set_level(config_get_int(cfg, NULL, "logging_level"));
 
-	if (STAILQ_EMPTY(&cfg_settings.interfaces)) {
+	assume_two_step = config_get_int(cfg, NULL, "assume_two_step");
+	sk_check_fupsync = config_get_int(cfg, NULL, "check_fup_sync");
+	sk_tx_timeout = config_get_int(cfg, NULL, "tx_timestamp_timeout");
+
+	ds->clockQuality.clockClass = config_get_int(cfg, NULL, "clockClass");
+	ds->clockQuality.clockAccuracy = config_get_int(cfg, NULL, "clockAccuracy");
+	ds->clockQuality.offsetScaledLogVariance =
+		config_get_int(cfg, NULL, "offsetScaledLogVariance");
+
+	dds->clock_desc.productDescription.max_symbols = 64;
+	dds->clock_desc.revisionData.max_symbols = 32;
+	dds->clock_desc.userDescription.max_symbols = 128;
+
+	tmp = config_get_string(cfg, NULL, "productDescription");
+	if (count_char(tmp, ';') != 2 ||
+	    static_ptp_text_set(&dds->clock_desc.productDescription, tmp)) {
+		fprintf(stderr, "invalid productDescription '%s'.\n", tmp);
+		goto out;
+	}
+	tmp = config_get_string(cfg, NULL, "revisionData");
+	if (count_char(tmp, ';') != 2 ||
+	    static_ptp_text_set(&dds->clock_desc.revisionData, tmp)) {
+		fprintf(stderr, "invalid revisionData '%s'.\n", tmp);
+		goto out;
+	}
+	tmp = config_get_string(cfg, NULL, "userDescription");
+	if (static_ptp_text_set(&dds->clock_desc.userDescription, tmp)) {
+		fprintf(stderr, "invalid userDescription '%s'.\n", tmp);
+		goto out;
+	}
+	tmp = config_get_string(cfg, NULL, "manufacturerIdentity");
+	if (OUI_LEN != sscanf(tmp, "%hhx:%hhx:%hhx", &oui[0], &oui[1], &oui[2])) {
+		fprintf(stderr, "invalid manufacturerIdentity '%s'.\n", tmp);
+		goto out;
+	}
+	memcpy(dds->clock_desc.manufacturerIdentity, oui, OUI_LEN);
+
+	ds->domainNumber = config_get_int(cfg, NULL, "domainNumber");
+
+	if (config_get_int(cfg, NULL, "slaveOnly")) {
+	    ds->flags |= DDS_SLAVE_ONLY;
+	    ds->clockQuality.clockClass = 248;
+	}
+	if (config_get_int(cfg, NULL, "twoStepFlag")) {
+	    ds->flags |= DDS_TWO_STEP_FLAG;
+	}
+	ds->priority1 = config_get_int(cfg, NULL, "priority1");
+	ds->priority2 = config_get_int(cfg, NULL, "priority2");
+
+	if (!config_get_int(cfg, NULL, "gmCapable") &&
+	    ds->flags & DDS_SLAVE_ONLY) {
+		fprintf(stderr,
+			"Cannot mix 1588 slaveOnly with 802.1AS !gmCapable.\n");
+		goto out;
+	}
+	if (!config_get_int(cfg, NULL, "gmCapable") ||
+	    ds->flags & DDS_SLAVE_ONLY) {
+		ds->clockQuality.clockClass = 255;
+	}
+	if (config_get_int(cfg, NULL, "clock_servo") == CLOCK_SERVO_NTPSHM) {
+		config_set_int(cfg, "kernel_leap", 0);
+		config_set_int(cfg, "sanity_freq_limit", 0);
+	}
+
+	if (STAILQ_EMPTY(&cfg->interfaces)) {
 		fprintf(stderr, "no interface specified\n");
 		usage(progname);
-		return -1;
+		goto out;
 	}
 
 	if (!(ds->flags & DDS_TWO_STEP_FLAG)) {
-		switch (*timestamping) {
+		switch (config_get_int(cfg, NULL, "time_stamping")) {
 		case TS_SOFTWARE:
 		case TS_LEGACY_HW:
 			fprintf(stderr, "one step is only possible "
 				"with hardware time stamping\n");
-			return -1;
+			goto out;
 		case TS_HARDWARE:
-			*timestamping = TS_ONESTEP;
+			if (config_set_int(cfg, "time_stamping", TS_ONESTEP))
+				goto out;
 			break;
 		case TS_ONESTEP:
 			break;
 		}
 	}
 
-	switch (*timestamping) {
+	switch (config_get_int(cfg, NULL, "time_stamping")) {
 	case TS_SOFTWARE:
 		required_modes |= SOF_TIMESTAMPING_TX_SOFTWARE |
 			SOF_TIMESTAMPING_RX_SOFTWARE |
@@ -336,27 +295,28 @@ int main(int argc, char *argv[])
 
 	/* Init interface configs and check whether timestamping mode is
 	 * supported. */
-	STAILQ_FOREACH(iface, &cfg_settings.interfaces, list) {
-		config_init_interface(iface, &cfg_settings);
+	STAILQ_FOREACH(iface, &cfg->interfaces, list) {
+		config_init_interface(iface, cfg);
 		if (iface->ts_info.valid &&
 		    ((iface->ts_info.so_timestamping & required_modes) != required_modes)) {
 			fprintf(stderr, "interface '%s' does not support "
 				        "requested timestamping mode.\n",
 				iface->name);
-			return -1;
+			goto out;
 		}
 	}
 
 	/* determine PHC Clock index */
-	iface = STAILQ_FIRST(&cfg_settings.interfaces);
-	if (cfg_settings.dds.free_running) {
+	iface = STAILQ_FIRST(&cfg->interfaces);
+	if (config_get_int(cfg, NULL, "free_running")) {
 		phc_index = -1;
-	} else if (*timestamping == TS_SOFTWARE || *timestamping == TS_LEGACY_HW) {
+	} else if (config_get_int(cfg, NULL, "time_stamping") == TS_SOFTWARE ||
+		   config_get_int(cfg, NULL, "time_stamping") == TS_LEGACY_HW) {
 		phc_index = -1;
 	} else if (req_phc) {
 		if (1 != sscanf(req_phc, "/dev/ptp%d", &phc_index)) {
 			fprintf(stderr, "bad ptp device string\n");
-			return -1;
+			goto out;
 		}
 	} else if (iface->ts_info.valid) {
 		phc_index = iface->ts_info.phc_index;
@@ -364,7 +324,7 @@ int main(int argc, char *argv[])
 		fprintf(stderr, "ptp device not specified and\n"
 			        "automatic determination is not\n"
 			        "supported. please specify ptp device\n");
-		return -1;
+		goto out;
 	}
 
 	if (phc_index >= 0) {
@@ -373,23 +333,24 @@ int main(int argc, char *argv[])
 
 	if (generate_clock_identity(&ds->clockIdentity, iface->name)) {
 		fprintf(stderr, "failed to generate a clock identity\n");
-		return -1;
+		goto out;
 	}
 
-	clock = clock_create(phc_index, &cfg_settings.interfaces,
-			     *timestamping, &cfg_settings.dds,
-			     cfg_settings.clock_servo);
+	clock = clock_create(cfg, phc_index, &cfg->interfaces, &ptp4l_dds);
 	if (!clock) {
 		fprintf(stderr, "failed to create a clock\n");
-		return -1;
+		goto out;
 	}
 
+	err = 0;
+
 	while (is_running()) {
 		if (clock_poll(clock))
 			break;
 	}
-
-	clock_destroy(clock);
-	config_destroy(&cfg_settings);
-	return 0;
+out:
+	if (clock)
+		clock_destroy(clock);
+	config_destroy(cfg);
+	return err;
 }
